/* VECTOR routine suite for variable-bit integer coding
 * Original vector and Elias coding Copyright (C) 1996 to Justin Zobel
 * This code, variable byte integers, and Golomb coding Copyright (C) 1996-8 
 * to Hugh Williams
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * See Copyright.txt for further details. */

/* Code for creating and accessing bitvectors.  They cannot be accessed
   until complete; once complete, they can only be extended by decoding,
   then recoding into a new vector. The code also assumes that the numbers
   to be encoded are at least one bit shorter than an integer. */

#include "def.h"

/* #define DEBUGGOLOMB             */
/* #define DEBUGELIAS */

int disaster = FALSE;
static int putnbits(BITVEC *,int,int);
static int getnbits(BITVEC *,int);
static int putbit(BITVEC *,int);
static int getbit(BITVEC *);

#ifdef DIAGS
	int totalvec = 0;
#endif

GOLHASH	golhashtable[GOLHASHTABLESIZE];

/* Number of bits written to any bitvector; used for consistency check */
static int bitcount;

void
initcmp()
{
    bitcount = 0;
}


/* Return length of vector */
int
veclen(BITVEC *vp)
{
    return(vp->len);
}


/* Before a vector can be read, it must be reset */
void
resetvec(BITVEC *vp)
{
    vp->last = -1;
    vp->pos = vp->bit = 0;
	vp->cur = vp->vector[0];
}

void
freevec(BITVEC *vp)
{
	free(vp->vector);
}

/* Initialising a vector before insertion of data */
void
initvec(BITVEC *vp, int initveclen)
{
    vp->vector = malloc(initveclen * sizeof(char));
    vp->size = initveclen;
    vp->last = -1;
    vp->bit = vp->pos = vp->len = vp->cur = 0;
}


/* Write the contents of a vector to file.  Assumes that padvec has
   been called. */
void
dumpvec(BITVEC *vp, FILE *fp)
{
#ifdef DEBUG 
    int i;

    resetvec(vp);
    fprintf(stderr,"[", i);
    while( (i=getrunlen(vp)) >=0 )
	fprintf(stderr," %d", i);
    fprintf(stderr,"] {", i);
    resetvec(vp);
    while( (i=getdelta(vp)) >=0 )
	fprintf(stderr," %d", i);
    fprintf(stderr,"}", i);
#endif  /* DEBUG */

    fwrite(&vp->len, sizeof(int), 1, fp);
    fwrite(vp->vector, sizeof(char), vp->len/8 + 1, fp);
}


/* Read a vector back from file */
/* Returns -1 on error */
int
readvec(BITVEC *vp, FILE *fp)
{
    int		len;

    fread(&vp->len, sizeof(int), 1, fp);
    vp->size = 4 * ( ( vp->len/8 + 8 ) / 4 ); /* word boundaries */
    len = vp->len/8 + 1;
    vp->vector = malloc(vp->size * sizeof(char));
    fread(vp->vector, sizeof(char), len, fp);
    resetvec(vp);
    return(0);
}


/* Put a unary number in a vector (unary means x is represented as x 0-bits, followed
   by a single 1-bit. Cute for small numbers */
int putunary(BITVEC *vp, int n)
{
	int ret;
	
    ret = putnbits(vp, 0, n);
    ret += putbit(vp, 1);
	
	return(ret);
}

int getunary(BITVEC *vp)
{
	int ret = 0;
	
	for(;getbit(vp) != 1;ret++);
	
	return(ret);
}

/* Add the number n to the vector, assuming run-length coding */
int
putrunlen(BITVEC *vp, int n)
{
    int		ret;

    if( n <= vp->last )
		return(0);		/* hack: don't add the same code twice */
    ret = putdelta(vp, n - vp->last);
    vp->last = n;
    return(ret);
}


/* Get a number from the vector, assuming run length coding */
int
getrunlen(BITVEC *vp)
{
    int		ret;

    ret = getdelta(vp);
    if( ret > 0 )
    {
		vp->last += ret;
		return(vp->last);
    }
    else
		return(-1);
}


/* Standard delta (Elias) */
int
putdelta(BITVEC *vp, int n)
{
    double	floor(), log10();
    int		mag;
    int		ret;

    mag = (int) floor(L2((double) n));
    
    ret = putgamma(vp, mag+1);
    ret += putnbits(vp, n, mag); /* don't output leftmost (ie, top) bit */
    
    return(ret);
}


/* Returns -1 on eof */
int
getdelta(BITVEC *vp)
{
    int		mag, val;

    mag = getgamma(vp) - 1;
    if( mag < 0 )
		return(-1);
    
    val = getnbits(vp, mag);
    
    if( val < 0 )
		return(-1);

    return( ( 1 << mag ) | val );
}


/* Standard gamma (Elias) */
int
putgamma(BITVEC *vp, int n)
{
    double	floor(), log10();
    int		mag;
    int		ret;

    mag = (int) floor(L2((double) n));
    
    ret = putnbits(vp, 0, mag);
    ret += putnbits(vp, n, mag+1);
	
    return(ret);
}


/* Returns -1 on eof */
int
getgamma(BITVEC *vp)
{
    int		b, mag, val;

    for( mag=0 ; ( b = getbit(vp) ) == 0 ; mag++ );

    if( b < 0 )
    {
		return(-1);
    }
    val = getnbits(vp, mag);
    if( val < 0 )
    {
		return(-1);
    }
    return( ( 1 << mag ) | val );
}

void
initgolhash()
{
	int x;
	for (x=0;x<GOLHASHTABLESIZE;x++)
		golhashtable[x].b = -1;
}

/* Standard Golomb */
/* x = value to be put in */
/* b = coding parameter */
int
putgolomb(BITVEC *vp, int x, int b)
{
    int		mag;
    int		ret;
    int     rem;
	int		q;
	int		d;

	/* Force +1 onto any value stored---to ensure non-zero values */
	x++;

	/* work out floor of (x - 1)/b */
    mag = (int) floor(((double)(x - 1))/((double)b));		
#ifdef DEBUGGOLOMB
	printf("-----------\nstoring: %d with b: %d, mag: %d\n", x,b,mag);
#endif	

	/* Store mag+1 in unary */
	/* Think about it: mag=3 means storing 001, hence this DOES work */
	ret = putnbits(vp, 0, mag);
	ret += putbit(vp, 1);
	
#ifdef DEBUGGOLOMB
	printf("stored mag of: %d\n", mag+1);
#endif

	/* What's the remainder? ie. x - (mag * b) */
	rem = x - (mag * b) - 1;
#ifdef DEBUGGOLOMB
	printf("rem: %d\n", rem);
#endif	
	
	/* Now work out the point at which coding requires floor(log(b))+1 bits */
	/* There are b remainders, so 2^k < b <= 2^(k+1) */
	/* Therefore, 2^k <= b and then k = log_2 b, so first k remainders */
	/* can be represented in floor(log_2 b) bits [phew] */
	
	q = (int) floor(L2((double) b));

	/* Now, work out the base value to use when we hit the toggle point (q) 		*/
	/* Since there are q+1 bits (max) used to represent a remainder, there 			*/
	/* can be 2^(q+1) such values using q+1 bits. Since there are b values 			*/
	/* we want to represent, if the remainder is less than 2^(q+1) - b, in other 	*/
	/* words if there are "spares", we can use q bits [phew again!]. 				*/	
	/* Geez, Golomb coding isn't fun.												*/
	/* So, d = 2^(q+1) - b		(jz/sara figured this, thanks)						*/
	d = (1 << (q+1)) - b; 

#ifdef DEBUGGOLOMB
	printf("toggle point q: %d with a d of: %d\n", q, d);
#endif	
	/* Can this remainder be represented in floor(log_2 b) bits? */
	if (rem < d)
		/* Store the value of remainder directly */
		ret += putnbits(vp, rem, q);	
	else
	{
#ifdef DEBUGGOLOMB
		printf("actually storing: %d in toggle+1 (%d) bits\n", d+rem, q+1);
#endif	
		/* Store the value of (d + remainder) in floor(log_2 b) + 1 bits */
		ret += putnbits(vp, d+rem, q+1);
	}
#ifdef DEBUGGOLOMB
	printf("stored: %d bits\n-----------\n", ret);
#endif	
	return(ret);	
}

/* Standard Golomb */
/* b = coding parameter */
int
getgolomb(BITVEC *vp, int b)
{
    int		mag;
	int		q, d;
	int		rem;
	int		ret;
	int		slot;
	GOLHASH	golhash;

	/* Get mag+1 in unary */
	/* By setting mag=0, we actually get mag [rather than mag+1, which */
	/* is what was stored] [phew!] */
    for( mag=0 ; getbit(vp) == 0 ; mag++ );
#ifdef DEBUGGOLOMB
	printf("-----------\ngot a mag: %d, meaning mag is: %d\n", mag+1, mag);
#endif

	/* Try and look up the log_2(b)... if not there, calculate and store */
	golhash = golhashtable[b % GOLHASHTABLESIZE];
	if (golhash.b == b)
		q = golhash.q;
	else
	{
		/* Now work out the toggle point for remainders as before */
		q = (int) floor(L2 ((double) b));
		slot = b % GOLHASHTABLESIZE;
		golhashtable[slot].b = b;
		golhashtable[slot].q = q;					   	 	
	}

#ifdef DEBUGGOLOMB
	printf("toggle: %d\n", q);
#endif

	/* Get the first q bits (we may need (q+1) actually) */
    rem = getnbits(vp, q);	

#ifdef DEBUGGOLOMB
	printf("rem: %d\n", rem);
#endif

	/* Now, work out the base value to use for the toggle point (q) */
	/* It's d=2^(q+1) - b           (jz/sara figured this, thanks)	*/
	/* See my notes in putgolomb() */
	d = (1 << (q+1)) - b; 

	/* Do we need to get an extra bit? */
	if (rem >= d)
	{
		/* If so: 
		   (1) shift rem left and OR with the bit we get 
	   	   (2) subtract d to get remainder */
		rem = ((rem << 1) | getbit(vp)) - d;

#ifdef DEBUGGOLOMB
	printf("new rem: %d\n", rem);
#endif
	}
	
	/* The value is the mag * b + remainder + 1 */
	ret =  mag * b + rem + 1;

#ifdef DEBUGGOLOMB
	printf("ret stored: %d", ret);
#endif

	/* Force -1 onto any value stored---used it putgolomb to ensure non-zero values */
	ret--;
	
#ifdef DEBUGGOLOMB
	printf("ret returned: %d\n-----------\n", ret);
#endif
	return(ret);	
}

/* Put n bits in vector, starting with nth from right, going to rightmost */
/* This is not efficient.  Should be recoded to insert several bits
   simultaneously ... but not really worth the effort ... */
int
putnbits(BITVEC *vp, int n, int num)
{
    int		i;
    int		ret;

    for( ret=0, i=num-1 ; i>=0 ; i-- )
		ret += putbit(vp, ( n >> i ) & 0x1 );
    return(ret);
}


static int masks[9] = { 0x0, 0x1, 0x3, 0x7, 0xf, 0x1f, 0x3f, 0x7f, 0xff };


/* Get n bits from vector */
/* Rather than calling getbit for each bit, shifts several bits at
   once if this would empty the current byte; only makes it one or
   two percent faster. */
int
getnbits(BITVEC *vp, int num)
{
    int		mask, shift, val, b;

    b = val = 0;

#if true
    for( shift = 8-vp->bit ; num >= shift ; num -= shift, shift = 8-vp->bit )
		/* copy out whole of cur */
    {
		if( 8*vp->pos + vp->bit >= vp->len)
		    return(-1);
		mask = masks[shift];
		val = ( val << shift ) | ( vp->cur & mask );
		vp->bit = 0;
		vp->pos++;
		vp->cur = vp->vector[vp->pos];
    }
#endif /* true */

	/* Get any remaining bits */
    for( ; num>0 && ( b = getbit(vp) ) >= 0 ; num-- )
		val = ( val << 1 ) | ( b & 0x1 );
#ifdef VECDEBUG
    if( val < 0 )
	printf("Panic in getnbits: overflow\n");
    if( b < 0 && vp->pos < vp->len/8 )
	printf("Early termination in getnbits.\n");
#endif /* VECDEBUG */
    return( ( b<0 ) ? -1 : val );
}


/* Put a bit into vector; bytes are filled left to right */
/* Because of the way bytes are filled, vectors can't be accessed until
   padvec has been called. */
int
putbit(BITVEC *vp, int b)
{
    bitcount++;

    vp->cur = ( vp->cur << 1 ) | ( b & 0x1 );
    vp->bit++;
    if( vp->bit == 8 ) /* go to next byte */
    {
	if( vp->pos >= vp->size )
	    expandvec(vp);

	if (disaster == FALSE)
	{
	vp->vector[vp->pos] = vp->cur & 0xff;
	vp->cur = 0;
	vp->pos++;
	vp->bit = 0;
    }
	}
    return(1);
}


/* Get a bit from a vector */
int
getbit(BITVEC *vp)
{
    int		b;

#ifdef DEBUGBIT
	printf("yep\n");
#endif

    if( 8*vp->pos + vp->bit >= vp->len )
		return(-1);

    b = ( vp->cur >> ( 7 - vp->bit ) ) & 0x1;
    vp->bit++;
    
    if( vp->bit == 8 )
    {
		vp->pos++;
		vp->bit = 0;
		vp->cur = vp->vector[vp->pos];
    }
    return(b);
}


/* Terminate a vector */
void
padvec(BITVEC *vp)
{
    if( vp->pos >= vp->size ) /* this is a bit yucky */
		expandvec(vp);
	if (disaster == FALSE)
	{
    vp->vector[vp->pos] = ( vp->cur << ( 8-vp->bit ) ) & 0xff;
    vp->len = 8*vp->pos + vp->bit;
	}
}


/* Enlarge vector */
void
expandvec(BITVEC *vp)
{
    char	*newvec;
    int		i;
    
	/* should use realloc here */
    if ((newvec = malloc( vp->size*2 * sizeof(char) )) == NULL)
    {
    	fprintf(stderr, "Error in expandvec malloc. Tried to allocate: %d bytes\n", vp->size*2 * sizeof(char));
		disaster = TRUE;
    }	
	else
	{
	for( i=0 ; i<vp->size ; i++ )
		newvec[i] = vp->vector[i];
	vp->size = vp->size*2;
	free(vp->vector);
	vp->vector = newvec;
	}
}
